unit Cards;

interface

uses Windows, Graphics, Classes, SysUtils, Controls, ExtCtrls, Types, Forms, IniFiles;

type
  TMast = (tmK, tmC, tmP, tmB);
  TTypeListCard = (tKoloda, tFreeCell, tHomeCell);

  PRGB_record = ^TRGB_record;
  TRGB_record = record
    B,G,R: Byte; // Blue, Green, Red
  end;

  TListCard = class;

  TCard = class
  private
    mast: TMast;
    nomer: 2..14;
    isInternalMoving: Boolean;
    x_start, y_start, x_aim, y_aim, delta_x, delta_y, dist_to_aim, count_step, nom_step: Integer;
    Pict: TBitMap;
    STRETCH_PIC: TBitMap;
    isNeedDraw: Boolean;
    fParent: TListCard;
    iWantShadowDraw: Boolean;
    function GetNextCard: TCard;
    function GetPrevCard: TCard;
    procedure RedrawKontur;
  public
    Left, Top: Integer;
    constructor Create(m: TMast; n: Byte);
    destructor Destroy; override;
    procedure SetIntervalToTopCards;
    procedure SetPosition(X,Y: Integer);
    procedure SetAimForMoving(X,Y: Integer);
    procedure MoveOnStep;
    procedure Draw(Canvas: TCanvas; DrawShadow: Boolean);
    function isMouseOnCard(x,y: Integer): boolean;
    function NeedRedraw: Boolean;
    procedure STRETCH_CARD;
  end;

  TListCard = class
  private
    ListOfCard: Tlist;
    LeftCoord, TopCoord: Integer;
    TypeList: TTypeListCard;
  public
    constructor Create;
    destructor Destroy; override;
    function GetCard(ind: Integer): TCard;
    function GetLastCard: TCard;
    function GetFirstCard: TCard;
    procedure Draw(Canvas: TCanvas; BaseCard: TCard); // BaseCard -      
    function HasCard(C: TCard): Boolean;
  end;

  TSoliter = class(TImage)
  private
    ArrayCards: array [1..52] of TCard;
    ArrayStolb: array [1..16] of TListCard;
    CardUnderMouse: TCard;
    x_down, y_down: Integer;
    nomer_active_st: Integer;
    is_moving: Boolean;
    LeftTimeToHideText: Integer;
    TEXT_info: string;
    isAutomaticMode: Boolean;
    HeWasInHome: TCard; //       
    UndoList: TStringList;
    isUndoMode: Boolean;
    VariantsMove: TStringList;
    LastVariantMove: string;
    isDblClick: Boolean;
    procedure SetText(txt: string);
    function GetIndexOfListCard(List: TListCard): Integer;
    function GetIndexOfCard(C: TCard): Integer;
    procedure TryAutomaticMove; //     
    procedure SaveMoveToUndoList(Card: TCard);
    procedure DropCard(Card: TCard);
    function CanDropCard(Card: TCard; NewListCard: TListCard): Boolean; //       ?
    function HasFreeCell(Card: TCard; ind: Integer): Boolean;
    procedure GoToTheNewPlace(Card: TCard; NewLC: TListCard); //      
    procedure InitVariantsMove; //        
  public
    function GetText: string;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure DblClick; override;
    procedure NewGame(NeedRandom: Boolean);
    procedure DrawGame;
    procedure NextStep;
    procedure TempProcedure(param: Integer);
    function NeedRedraw: Boolean;
    procedure DoUndo;
    function CanUndo: Boolean;
    function GetCountMoves: Integer;
    procedure ShowVariantHoda; //    
    procedure LoadParamFromIniFile;
    procedure SaveParamToIniFile;    
    procedure ChangeConturColor;
    procedure RecalcNewSizeGame; //       
    procedure InitAllStolbec(isFirst: Boolean);
  end;

var
  CARDS_BITMAP       : TBitmap; // 
  FON_BITMAP         : TBitmap; // 
  PLACE_UP           : TBitmap; //    
  PLACE_DOWN         : TBitmap; //    

  PLACE_UP_ROUND     : TBitmap; //       
  PLACE_DOWN_ROUND   : TBitmap; //       

  BIG_FON_FOR_STRETCH: TBitmap; //  ,      FON_BITMAP

  GIRL_PICTURE       : TBitmap; // 
  LOGO               : TBitmap; //        
 
  COLOR_FROM, COLOR_AIM: TColor; //  
  FON_COLOR: TColor; //  
  NEED_SHOW_FREE_CELL: Boolean; // 
  SPEED_CARD: Integer; //    

  APPLICATION_PATH  : string;  
  CARD_WIDTH, CARD_HEIGHT: Integer; //     !!!

procedure DrawOreolRect(Canvas: TCanvas; Left, Top, Right, Bottom: Integer; ColorRect: TColor);  
procedure TransformBitMap(Color: TColor; BITMAP: TBitMap);
  
implementation

const

  _WidthCard = 114;                   //   
  _HeightCard = 154;                 //   
  _SizeShadow = _WidthCard div 8;   //  
  _DeltaShadow = _SizeShadow div 2;//          (  )
  _SizeOreol = _WidthCard div 5;  //  
  _RadiusCard = 3;               //   
  _ShadowLevel = 0.5;           //       0..1 (0 - , 1 - )
  _TimeShowText = 500;         //  [ ]   
  _nitro_speed = 2.5;         //      free cell

  SMOOTH = 1.2; //                     

  TRANSPARENT_COLOR = clLime; // ,   

  absX1 = _SizeShadow;                                              //         (abs - AlphaBlendShadow)
  absY1 = _HeightCard - _SizeShadow - _RadiusCard + _DeltaShadow;   //   
  absX2 = _WidthCard + _DeltaShadow;                                //   
  absY2 = _SizeShadow + _RadiusCard;                                //   

  ablX  = _WidthCard + 2*_SizeOreol;      //    
  ablY  = _SizeOreol + _RadiusCard;       //     (  )
  
  _PROCENT_AUTO_MOVE = 30;  //          

  _txt_WrongHod = '  ';
  _txt_FreeCell = ' : ';
  _txt_MoveCard = ' : ';
  _txt_MaxMoveC = '   : ';
  _txt_DontMove = '    ';
  _txt_IsByzzzi = ',  ';
  _txt_WrongMas = ',     ';
  _txt_Tuz_Need = ',     ';

var
  AlphaBlendShadow1 : array [0..absX1-1, 0..absY1-1] of Real;  //    ( )
  AlphaBlendShadow2 : array [0..absX2-1, 0..absY2-1] of Real;  //    ( )
  AlphaBlendLighting: array [0..ablX -1, 0..ablY -1] of Real;  //   
  Shadow1, Shadow2  : TBitMap; //     
  RideTheLighting   : TBitMap; //     
  move_with_delay   : Boolean; // True      MoveCardWithDelay

//  ( )    //

  _MinZazor   : Integer = _HeightCard div 6;   //      
  _MaxZazor   : Integer = _HeightCard div 3;  //          ( )
  _LeftZazor  : Integer = 15;                //        
  _TopZazor   : Integer = 180;              //        
  _TopTopZazor: Integer = 15;              //        
  _Left1Zazor : Integer = 15;             //         
  _Left2Zazor : Integer = 536;           //         
  _TopShagKart: Integer = 118;          //   
  _ShagKart   : Integer = 125;         //     
  
                       
type
  TTypeColor = (tcMask, tcShadow, tcPicture);

function RGB(r, g, b: Byte): TColor;
begin
  Result := r + g shl 8 + b shl 16;
end;

function FadeToBlack(ColorRGB: Byte; HowMuch: Extended): Byte; // HowMuch 0..1, 0 - , 1 -  
begin
  Result := Round(ColorRGB*HowMuch);
end;

function FadeColorToBlack(Color: TColor; HowMuch: Extended): TColor; // HowMuch 0..1, 0 - , 1 -  
var
  r,g,b: Byte;
begin
  r := (Color and $0000FF);
  g := (Color and $00FF00) shr 8;
  b := (Color and $FF0000) shr 16;
  Result := RGB(FadeToBlack(r, HowMuch), FadeToBlack(g, HowMuch), FadeToBlack(b, HowMuch));
end;

function MixColor_50_50(Color1, Color2: TColor): TColor; //    50/50
var
  r1,g1,b1: Byte;
  r2,g2,b2: Byte;
begin
  r1 := (Color1 and $0000FF);
  g1 := (Color1 and $00FF00) shr 8;
  b1 := (Color1 and $FF0000) shr 16;
  r2 := (Color2 and $0000FF);
  g2 := (Color2 and $00FF00) shr 8;
  b2 := (Color2 and $FF0000) shr 16;
  Result := RGB((r1+r2) div 2, (g1+g2) div 2, (b1+b2) div 2);
end;

function MixColors(Color1, Color2: Byte; A1,A2: Real): Byte; // A1+A2 = 1
begin
  Result := Round(A1*Color1 + A2*Color2);
end;

procedure GetCardPic(var PictCard: TBitMap; m: TMast; n: Byte); //    
var 
  dx,dy: Integer;
begin
  dx := (n-2)*_WidthCard;
  case m of
    tmK: dy := _HeightCard*1;
    tmC: dy := _HeightCard*2;
    tmP: dy := _HeightCard*3;
  else
    dy := 0;
  end;
  PictCard.Canvas.CopyRect(PictCard.Canvas.ClipRect, CARDS_BITMAP.Canvas, RECT(dx,dy,_WidthCard+dx,_HeightCard+dy));
end;

procedure RoundPic(var PictCard: TBitMap; Radius: Integer); //  
var
  RadBitMap: TBitMap;
begin
  RadBitMap := TBitmap.Create;
  
  RadBitMap.Width := PictCard.Width;
  RadBitMap.Height := PictCard.Height;
  RadBitMap.Canvas.Brush.Color := TRANSPARENT_COLOR;
  RadBitMap.Canvas.FillRect(RadBitMap.Canvas.ClipRect);
  RadBitMap.Canvas.Brush.Color := clBlack;  
  RadBitMap.Canvas.Pen.Color := TRANSPARENT_COLOR;
  RadBitMap.Canvas.RoundRect(0,0, RadBitMap.Width, RadBitMap.Height, Radius*2, Radius*2);

  RadBitMap.TransparentColor := clBlack;
  RadBitMap.Transparent := True;
  PictCard.Canvas.Draw(0,0, RadBitMap);
  
  RadBitMap.Free;  

  PictCard.TransparentColor := TRANSPARENT_COLOR;
  PictCard.Transparent := True;
end;

constructor TCard.Create(m: TMast; n: Byte);
begin
  inherited Create;
  mast := m;
  nomer := n;
  Pict := TBitmap.Create;
  Pict.PixelFormat := pf24bit;  
  isInternalMoving := False;
  isNeedDraw := False;
  iWantShadowDraw := False;
  fParent := nil;

  Pict.Width := _WidthCard;
  Pict.Height := _HeightCard;

  GetCardPic(Pict, m,n);   //       

  Left := -Pict.Width;
  Top := FON_BITMAP.Height;

  STRETCH_PIC := TBitmap.Create;
  STRETCH_PIC.PixelFormat := pf24bit;  
  STRETCH_PIC.Transparent := True;
  STRETCH_CARD;
end;

function StretchPicture(PicToStretch, Source: TBitmap; NewWidth, NewHeight: Integer): Boolean;
begin
  Result := False;
  if (PicToStretch.Width = NewWidth) and (PicToStretch.Height = NewHeight) then Exit;
  PicToStretch.Width := NewWidth;
  PicToStretch.Height := NewHeight;
  SetStretchBltMode(PicToStretch.Canvas.Handle, HALFTONE);
  StretchBlt(PicToStretch.Canvas.Handle, 0, 0, NewWidth,NewHeight, Source.Canvas.Handle, 0, 0, Source.Width, Source.Height, SRCCOPY); 
  Result := True;
end;

procedure TCard.STRETCH_CARD;
begin
  if StretchPicture(STRETCH_PIC, Pict, CARD_WIDTH, CARD_HEIGHT) then
    begin
      RoundPic(STRETCH_PIC, _RadiusCard);
      RedrawKontur;
    end;
end;

procedure TCard.RedrawKontur;
begin
  STRETCH_PIC.Canvas.Pen.Color := MixColor_50_50(FON_COLOR, clSilver);
  STRETCH_PIC.Canvas.Pen.Color := MixColor_50_50(STRETCH_PIC.Canvas.Pen.Color, clSilver);  
  STRETCH_PIC.Canvas.Brush.Style := bsClear;
  STRETCH_PIC.Canvas.RoundRect(0,0, CARD_WIDTH, CARD_HEIGHT, _RadiusCard*2, _RadiusCard*2);
end;

destructor TCard.Destroy;
begin
  Pict.Free;
  STRETCH_PIC.Free;
  inherited Destroy;
end;

function TCard.GetNextCard: TCard;
var
  ind: Integer;
begin
  Result := nil;
  if fParent <> nil then
    begin
      ind := fParent.ListOfCard.IndexOf(Self);
      if fParent.ListOfCard.Count - 1 > ind then
        Result := fParent.GetCard(ind + 1);
    end;
end;

function TCard.GetPrevCard: TCard;
var
  ind: Integer;
begin
  Result := nil;
  if fParent <> nil then
    begin
      ind := fParent.ListOfCard.IndexOf(Self);
      if ind > 0 then
        Result := fParent.GetCard(ind - 1);
    end;
end;

procedure TCard.SetIntervalToTopCards;
var
  NC: TCard;
  MZ: Integer;
begin
  NC := GetNextCard;
  MZ := _MinZazor;
  if fParent <> nil then if fParent.TypeList = tHomeCell then MZ := 0;
  if NC <> nil then NC.SetAimForMoving(Left, Top + MZ);
end;

procedure TCard.SetPosition(X,Y: Integer);
var
  NC: TCard;
  MZ: Integer;
begin
  Left := X;
  Top := Y;
  isNeedDraw := False;
  if (fParent = nil) or (fParent.TypeList = tHomeCell) then
    MZ := 0
  else
    MZ := _MinZazor;
  NC := GetNextCard;
  if NC <> nil then
    begin
      NC.SetAimForMoving(X, Y + MZ);
      NC.iWantShadowDraw := iWantShadowDraw;
    end;
end;

procedure TCard.SetAimForMoving(X,Y: Integer);
begin
  isInternalMoving := True;
  x_start := Left;
  y_start := Top;
  x_aim := X;
  y_aim := Y;
  delta_x := x_aim - x_start;
  delta_y := y_aim - y_start;
  dist_to_aim := ROUND(sqrt(sqr(delta_x) + sqr(delta_y)));
  count_step := dist_to_aim div SPEED_CARD + 3;  // 3   
  if dist_to_aim < _SizeShadow then Dec(count_step);        //      
  if dist_to_aim < _SizeShadow div 2 then Dec(count_step);  //      
  nom_step := 0;
end;

procedure TCard.MoveOnStep;
var
  xxx,yyy, dx,dy: Integer;
  L: Extended;
  PrevC: TCard;
begin
  if isInternalMoving then
    begin
      Inc(nom_step);
      if nom_step = count_step then // 
        begin
          SetPosition(x_aim, y_aim);
          isInternalMoving := False;
          isNeedDraw := True;
          PrevC := GetPrevCard;
          if PrevC <> nil then
            iWantShadowDraw := PrevC.iWantShadowDraw //  
          else
            iWantShadowDraw := False;
        end
      else
        begin
          L := nom_step/count_step;
          dx := Round(L * delta_x);
          dy := Round(L * delta_y);
          xxx := dx + x_start;
          yyy := dy + y_start;
          SetPosition(xxx,yyy);
        end;
    end;
end;

procedure TCard.Draw(Canvas: TCanvas; DrawShadow: Boolean);
var
  x,y,dx,dw: Integer;
  p: PRGB_record;
  shd: Extended;
  RectSh: TRect;
begin
  if DrawShadow or iWantShadowDraw then
    begin
      //  1:
      RectSh := Rect(Left + CARD_WIDTH, Top + _SizeShadow - _DeltaShadow, Left + CARD_WIDTH + Shadow1.Width, Top + _SizeShadow - _DeltaShadow + Shadow1.Height);
      Shadow1.Canvas.CopyRect(Shadow1.Canvas.ClipRect, Canvas, RectSh);
      for y := 0 to Shadow1.Height - 1 - (_HeightCard-CARD_HEIGHT) do
        begin
          p := Shadow1.ScanLine[y];
          for x := 0 to Shadow1.Width - 1 do
            begin
              shd := AlphaBlendShadow1[x,y];
              p^.R := FadeToBlack(p^.R, shd);
              p^.G := FadeToBlack(p^.G, shd);
              p^.B := FadeToBlack(p^.B, shd);
              Inc(p);
            end;
        end;
      Canvas.Draw(RectSh.Left, RectSh.Top, Shadow1);
      //  2:
      RectSh := Rect(Left + _SizeShadow - _DeltaShadow, Top + CARD_HEIGHT - _RadiusCard, Left + _SizeShadow - _DeltaShadow + Shadow2.Width,  Top + CARD_HEIGHT - _RadiusCard + Shadow2.Height);
      Shadow2.Canvas.CopyRect(Shadow2.Canvas.ClipRect, Canvas, RectSh);
      dw := _WidthCard-CARD_WIDTH;
      for y := 0 to Shadow2.Height - 1 do
        begin
          p := Shadow2.ScanLine[y];
          dx := 0;
          for x := 0 to Shadow2.Width - 1 - dw do
            begin
              if (dx = 0) and (x > _SizeShadow+_RadiusCard div 2) then dx := dw;
              shd := AlphaBlendShadow2[x + dx ,y];
              p^.R := FadeToBlack(p^.R, shd);
              p^.G := FadeToBlack(p^.G, shd);
              p^.B := FadeToBlack(p^.B, shd);
              Inc(p);
            end;
        end;
      Canvas.Draw(RectSh.Left, RectSh.Top, Shadow2);
    end;
  //  
  Canvas.Draw(Left, Top, STRETCH_PIC); {}
  isNeedDraw := False;
end;

function TCard.isMouseOnCard(x,y: Integer): boolean;
begin
  Result := (x > Left) and (x < Left + CARD_WIDTH) and (y > Top) and (y < Top + CARD_HEIGHT);
end;

function TCard.NeedRedraw: Boolean;
begin
  Result := isNeedDraw or isInternalMoving;
end;

function isFreiendlyCards(C1,C2: TCard): Boolean; //     
var
  isRed1, isRed2: Boolean;
begin
  Result := (C1.nomer - C2.nomer = 1) and (C1.nomer <> 14); // ..   1,       
  if Result then
    begin
      isRed1 := (C1.mast = tmC) or (C1.mast = tmB);
      isRed2 := (C2.mast = tmC) or (C2.mast = tmB);
      Result := (not isRed1 and isRed2) or (isRed1 and not isRed2); //   
    end;
end;

function isFrendlyHome(C1,C2: TCard): Boolean; //     
begin
  Result := C1.mast = C2.mast; // ..  
  if Result then
    Result := ((C1.nomer = 14) and (C2.nomer = 2)) or //    2-
              (C2.nomer - C1.nomer = 1); //   
end;

{TListCard}

constructor TListCard.Create;
begin
  inherited Create;
  ListOfCard := TList.Create;
end;

destructor TListCard.Destroy;
var
  i: Integer;
  C: TCard;
begin
  for i := 0 to ListOfCard.Count - 1 do
    begin
      C := GetCard(i);
      if Assigned(C) then C.Free;
    end;
  inherited Destroy;
end;

function TListCard.GetCard(ind: Integer): TCard;
begin
  try
    Result := TCard(ListOfCard[ind]);
  except
    Result := nil;
  end;
end;

function TListCard.GetLastCard: TCard;
begin
  if ListOfCard.Count > 0 then
    Result := GetCard(ListOfCard.Count-1)
  else
    Result := nil;
end;

function TListCard.GetFirstCard: TCard;
begin
  if ListOfCard.Count <> 0 then
    Result := GetCard(0)
  else
    Result := nil;
end;

procedure TListCard.Draw(Canvas: TCanvas; BaseCard: TCard);
var
  i: Integer;
  C: TCard;
  DrSh: Boolean;
begin
  DrSh := False;
  for i := 0 to ListOfCard.Count - 1 do
    begin
      C := GetCard(i);
      if not DrSh then if BaseCard = C then DrSh := True;
      C.Draw(Canvas, DrSh);
    end;
end;

function TListCard.HasCard(C: TCard): Boolean;
begin
  Result := ListOfCard.IndexOf(C) > -1;
end;

{TSoliter}

constructor TSoliter.Create(AOwner: TComponent);
var
  i: Integer;
begin
  inherited Create(AOwner);
  LastVariantMove := '';
  UndoList := TStringList.Create;
  VariantsMove := TStringList.Create;
  isUndoMode := False;
  NEED_SHOW_FREE_CELL := False;
  COLOR_FROM := clAqua; 
  COLOR_AIM  := clYellow; 
  nomer_active_st := 0;
  LeftTimeToHideText := 0;
  TEXT_info := '';
  FON_COLOR := clRed;
  SPEED_CARD := 50;
  isDblClick := False;
  Align := alClient;
  for i := 1 to 52 do
    case (i-1) div 13 of
      0: ArrayCards[i] := TCard.Create(tmK, (i-1) mod 13 + 2);
      1: ArrayCards[i] := TCard.Create(tmC, (i-1) mod 13 + 2);
      2: ArrayCards[i] := TCard.Create(tmP, (i-1) mod 13 + 2);
      3: ArrayCards[i] := TCard.Create(tmB, (i-1) mod 13 + 2);
    end;
  for i := 1 to 16 do ArrayStolb[i] := TListCard.Create;
  InitAllStolbec(True);
end;

destructor TSoliter.Destroy;
var
  i: Integer;
begin
  UndoList.Free;
  VariantsMove.Free;
  for i := 1 to 16 do ArrayStolb[i].Free;
  inherited Destroy;
end;

function CanStartMoveCard(MC: TCard): Boolean;
var
  MovingList: TListCard;
  i: Integer;
  C1,C2: TCard;
begin
  Result := True; //  1 
  MovingList := TListCard.Create;
  MovingList.ListOfCard.Add(MC);  //     
  while MC.GetNextCard <> nil do
    begin
      MC := MC.GetNextCard;
      MovingList.ListOfCard.Add(MC);
    end;
  for i := 1 to MovingList.ListOfCard.Count - 1 do
    begin
      C1 := MovingList.GetCard(i-1);
      C2 := MovingList.GetCard(i);
      Result := isFreiendlyCards(C1,C2);
      if not Result then Break;
    end;
  MovingList.ListOfCard.Clear;
  MovingList.Free;
end;

procedure MoveCardWithDelay(Card: TCard; X,Y: Integer);
begin
  move_with_delay := True;
  Card.iWantShadowDraw := True; //  
  Card.SetAimForMoving(X,Y);    //   
  while Card.iWantShadowDraw do Application.ProcessMessages; // ,  
  move_with_delay := False;
end;

procedure TSoliter.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  i: Integer;
  C: TCard;
begin
  inherited MouseDown(Button, Shift, X, Y);
  SetText('');
  if move_with_delay then Exit; //    
  is_moving := False;
  CardUnderMouse := nil;
  nomer_active_st := 0;
  for i := 1 to 16 do //       
    begin
      C := ArrayStolb[i].GetLastCard; //     
      while C <> nil do
        begin
          if C.isMouseOnCard(X,Y) then //   
            begin
              CardUnderMouse := C; //  
              nomer_active_st := i; //    
              Break; //    WHILE
            end;
          C := C.GetPrevCard; //  
        end;
      if CardUnderMouse <> nil then Break; //    FOR
    end;
  if CardUnderMouse <> nil then //    
    if Button = mbLeft then //  
      if CanStartMoveCard(CardUnderMouse) then //    
        begin
          is_moving := True;
          Mouse.CursorPos := Point(Mouse.CursorPos.X - _SizeShadow div 2, Mouse.CursorPos.Y - _SizeShadow div 2); // 
          x_down := X - CardUnderMouse.Left;
          y_down := Y - CardUnderMouse.Top;
          if nomer_active_st in [13..16] then //   
            HeWasInHome := CardUnderMouse //        
          else
            HeWasInHome := nil;
        end
      else
        begin
          SetText(_txt_DontMove);
          C := CardUnderMouse;
          CardUnderMouse := nil; //  ,   ...          
          {}
          MoveCardWithDelay(C, C.Left - _SizeShadow, C.Top);
          MoveCardWithDelay(C, C.Left + _SizeShadow*2, C.Top);
          MoveCardWithDelay(C, C.Left - _SizeShadow, C.Top);
          {}
        end
    else //     
      if CardUnderMouse.GetNextCard <> nil then //    
        CardUnderMouse.GetNextCard.SetAimForMoving(CardUnderMouse.Left, CardUnderMouse.Top + _MaxZazor) //   
      else
        CardUnderMouse := nil;
end;

procedure TSoliter.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  inherited MouseMove(Shift, X, Y);
  if (CardUnderMouse <> nil) and is_moving then
    CardUnderMouse.SetPosition(X - x_down, Y - y_down);
end;

procedure TSoliter.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);

  procedure TryMoveCardInUp(MovedCard: TCard);
  var
    i: Integer;
    Res: Boolean;
  begin
    Res := False;
    isAutomaticMode := True;
    for i := 13 to 16 {home} do
      if CanDropCard(MovedCard, ArrayStolb[i]) then //    
        begin //  
          GoToTheNewPlace(MovedCard, ArrayStolb[i]);
          Res := True;
          Break;
        end;
    if not Res then
      for i := 9 to 12 {left} do
        if CanDropCard(MovedCard, ArrayStolb[i]) then //    
          begin //  
            GoToTheNewPlace(MovedCard, ArrayStolb[i]);
            Break;
          end;
    isAutomaticMode := False;          
  end;

begin
  inherited MouseUp(Button, Shift, X, Y);
  if move_with_delay then Exit; //      
  if CardUnderMouse <> nil then
    if Button <> mbLeft then
      CardUnderMouse.SetIntervalToTopCards
    else
      begin
        Mouse.CursorPos := Point(Mouse.CursorPos.X + _SizeShadow div 2, Mouse.CursorPos.Y + _SizeShadow div 2);      
        DropCard(CardUnderMouse);        
        if isDblClick and (CardUnderMouse.GetNextCard = nil) then  //       ..
          begin
            TryMoveCardInUp(CardUnderMouse);
            TryAutomaticMove; //   
            InitVariantsMove; //   
          end;
      end;
  CardUnderMouse := nil;
  isDblClick := False;
end;

procedure TSoliter.DblClick;
begin
  isDblClick := True;
end;

procedure TSoliter.InitAllStolbec(isFirst: Boolean);

  procedure InitStolbec(IND, FirstIndex, LastIndex: Integer; TypeKol: TTypeListCard; NeedShadow: Boolean);  //    
  var
    i: Integer;
  begin
    ArrayStolb[IND].ListOfCard.Clear;
    ArrayStolb[IND].TypeList := TypeKol;
    case TypeKol of
      tKoloda: begin
                 ArrayStolb[IND].LeftCoord := _LeftZazor + (IND-1)*_ShagKart;
                 ArrayStolb[IND].TopCoord := _TopZazor;
               end;
      tFreeCell: begin
                   ArrayStolb[IND].LeftCoord := _Left1Zazor + (IND-9)*_TopShagKart;
                   ArrayStolb[IND].TopCoord := _TopTopZazor;
                 end;
      tHomeCell: begin
                   ArrayStolb[IND].LeftCoord := _Left2Zazor + (IND-13)*_TopShagKart;
                   ArrayStolb[IND].TopCoord := _TopTopZazor;
                 end;
    end;
    if (FirstIndex = -1) or not NeedShadow then Exit;
    for i := FirstIndex to LastIndex do
      begin
        ArrayStolb[IND].ListOfCard.Add(ArrayCards[i]);
        ArrayCards[i].fParent := ArrayStolb[IND];
        ArrayCards[i].iWantShadowDraw := NeedShadow; //     
      end;
  end;

begin
  InitStolbec(1, 01,07, tKoloda, not isFirst);
  InitStolbec(2, 08,14, tKoloda, not isFirst);
  InitStolbec(3, 15,21, tKoloda, not isFirst);
  InitStolbec(4, 22,28, tKoloda, not isFirst);
  InitStolbec(5, 29,34, tKoloda, not isFirst);
  InitStolbec(6, 35,40, tKoloda, not isFirst);
  InitStolbec(7, 41,46, tKoloda, not isFirst);
  InitStolbec(8, 47,52, tKoloda, not isFirst);

  InitStolbec(09, -1,0, tFreeCell, not isFirst);
  InitStolbec(10, -1,0, tFreeCell, not isFirst);
  InitStolbec(11, -1,0, tFreeCell, not isFirst);
  InitStolbec(12, -1,0, tFreeCell, not isFirst);

  InitStolbec(13, -1,0, tHomeCell, not isFirst);
  InitStolbec(14, -1,0, tHomeCell, not isFirst);
  InitStolbec(15, -1,0, tHomeCell, not isFirst);
  InitStolbec(16, -1,0, tHomeCell, not isFirst);
end;

procedure TSoliter.NewGame(NeedRandom: Boolean);

  procedure Obmen(var A,B: TCard);
  var
    C: TCard;
  begin
    C := A;
    A := B;
    B := C;
  end;

var
  i: Integer;

begin
  if NeedRandom then //   ( ) else ( )
    begin
      Randomize;
      for i := 1 to 52*52 do //     52 
        Obmen(ArrayCards[Random(52)+1], ArrayCards[Random(52)+1]);   //  0 <= Random(52) < 52, .. [0..51]
    end;   

  InitAllStolbec(False);  
    
  for i := 1 to 8 do
    ArrayStolb[i].GetFirstCard.SetAimForMoving(ArrayStolb[i].LeftCoord, ArrayStolb[i].TopCoord);

  UndoList.Clear; //   
  LastVariantMove := '';
  SetText('');
  InitVariantsMove; //    
end;

procedure TSoliter.DrawGame;
var
  i: Integer;
  c: TCard;
begin
  if is_moving then c := CardUnderMouse else c := nil;
  Picture.Bitmap.Assign(FON_BITMAP); //     Canvas.Draw(0,0, FON_BITMAP);
  for i := 1 to 16 do
    if i <> nomer_active_st then ArrayStolb[i].Draw(Canvas, c);
  if nomer_active_st <> 0 then ArrayStolb[nomer_active_st].Draw(Canvas, c); //   
end;

procedure TSoliter.NextStep;
var
  i: Integer;
begin
  for i := 1 to 52 do ArrayCards[i].MoveOnStep;
  if NeedRedraw then DrawGame;
  if LeftTimeToHideText > 0 then Dec(LeftTimeToHideText);
end;

procedure TSoliter.TempProcedure(param: Integer);
begin
  case param of
    0: MoveCardWithDelay(CardUnderMouse,10,10);
  end;
end;

function TSoliter.NeedRedraw: Boolean;
var
  i: Integer;
begin
  Result := CardUnderMouse <> nil;
  if Result then Exit;
  for i := 1 to 52 do
    if ArrayCards[i].NeedRedraw then
      begin
        Result := True;
        Exit;
      end;
end;

procedure TSoliter.DropCard(Card: TCard);
var
  i, ind: Integer;
  DeltaX: Integer;
  C: TCard;
begin
  C := Card;
  while C <> nil do
    begin
      C.iWantShadowDraw := True; //     (iWantShadowDraw := False   MoveOnStep)
      C := C.GetNextCard;
    end;

  ind := 0;
  for i := 1 to 16 do //      16 ""
    if i <> nomer_active_st then
      begin
        if i < 9 then DeltaX := _ShagKart div 2 + 2 else DeltaX := _TopShagKart div 2 + 2;
        if (abs(ArrayStolb[i].LeftCoord - Card.Left) < DeltaX) and
          ( ((_TopTopZazor + CARD_HEIGHT < Card.Top) and (i < 9)) OR //   
          ((Card.Top < _TopTopZazor + CARD_HEIGHT) and (i > 8) and (Card.GetNextCard = nil)) ) then //   
          begin
            ind := i;
            Break;
          end;
      end;
  if (ind > 0) and CanDropCard(Card, ArrayStolb[ind]) and HasFreeCell(Card, ind) then // ecm   ""   ind         
    begin
      GoToTheNewPlace(Card, ArrayStolb[ind]);
      nomer_active_st := ind; //       
      TryAutomaticMove; //   
      InitVariantsMove; //   
    end
  else  //       ,    :
    if ArrayStolb[nomer_active_st].GetFirstCard <> Card then //     
      ArrayStolb[nomer_active_st].GetFirstCard.SetIntervalToTopCards //    
    else //   ,   Card,    
      Card.SetAimForMoving(ArrayStolb[nomer_active_st].LeftCoord, ArrayStolb[nomer_active_st].TopCoord); //    
  is_moving := False;    
end;

function TSoliter.CanDropCard(Card: TCard; NewListCard: TListCard): Boolean; //       ?
var
  onCard: TCard;
begin
  Result := False;
  if Card = nil then Exit;
  onCard := NewListCard.GetLastCard;
  if onCard = nil then //   
    begin
      Result := True; //    
      if NewListCard.TypeList = tHomeCell then Result := Card.nomer = 14; //  
      if not Result then SetText(_txt_WrongHod + _txt_Tuz_Need);
      Exit;
    end;
  case NewListCard.TypeList of //    
    tKoloda:   Result := isFreiendlyCards(onCard, Card);
    tFreeCell: Result := FALSE; //     
    tHomeCell: Result := isFrendlyHome(onCard, Card);
  end;
  SetText('');
  if not Result then
    case NewListCard.TypeList of
      tKoloda, tHomeCell: SetText(_txt_WrongHod + _txt_WrongMas);
      tFreeCell: SetText(_txt_WrongHod + _txt_IsByzzzi);
    end;
end;


function TSoliter.HasFreeCell(Card: TCard; ind: Integer): Boolean;
var
  i, CountFreeCell, CountMoveCards: Integer;
  NC: TCard;
begin
  if ind > 12 then //   
    begin
      Result := True;
      SetText('');
      Exit;
    end;
  CountFreeCell := 0;
  for i := 1 to 12 do {8+4}
    if ArrayStolb[i].GetFirstCard = nil then
      Inc(CountFreeCell);
  CountMoveCards := 0;
  NC := Card;
  while NC <> nil do
    begin
      Inc(CountMoveCards);
      NC := NC.GetNextCard;
    end;
  if ArrayStolb[ind].ListOfCard.Count = 0 then Dec(CountFreeCell); //       
  Result := CountFreeCell - CountMoveCards >= -1;
  if not Result then
    SetText(_txt_FreeCell + IntToStr(CountFreeCell) + ';  ' +
            _txt_MoveCard + IntToStr(CountMoveCards) + ';  ' +
            _txt_MaxMoveC + IntToStr(CountFreeCell + 1) + ';')
  else
    SetText('');
end;

procedure TSoliter.GoToTheNewPlace(Card: TCard; NewLC: TListCard);
var
  MovingList, ListFC: TListCard;
  C: TCard;
  i, MZ: Integer;
  OldParentList: TListCard;
  HomePlace, NewPlace: TPoint;
begin

  if not isUndoMode then SaveMoveToUndoList(Card); //   ,       

  CardUnderMouse := nil;
  C := Card.GetPrevCard;
  if C <> nil then HomePlace := Point(C.Left, C.Top + _MinZazor) else HomePlace := Point(Card.fParent.LeftCoord, Card.fParent.TopCoord);
  C := NewLC.GetLastCard;
  if NewLC.TypeList = tHomeCell then MZ := 0 else MZ := _MinZazor;
  if C <> nil then NewPlace := Point(C.Left, C.Top + MZ) else NewPlace := Point(NewLC.LeftCoord, NewLC.TopCoord);
  MovingList := TListCard.Create;
  C := Card;
  MovingList.ListOfCard.Add(C);  //     
  while C.GetNextCard <> nil do
    begin
      C := C.GetNextCard;
      MovingList.ListOfCard.Add(C);
    end;
  OldParentList := Card.fParent;
  for i := 0 to MovingList.ListOfCard.Count - 1 do
    begin
      C := MovingList.GetCard(i);
      NewLC.ListOfCard.Add(C);  //    
      OldParentList.ListOfCard.Remove(C); //     
      C.fParent := NewLC;
    end;
  nomer_active_st := GetIndexOfListCard(NewLC); //   
  if MovingList.ListOfCard.Count = 1 then //   
    MoveCardWithDelay(Card, NewPlace.X, NewPlace.Y) //   
  else //       FreeCell
    begin // start FREE_CELL
      if isUndoMode or not NEED_SHOW_FREE_CELL then //   Undo   FreeCell
        begin
          Card.SetAimForMoving(NewPlace.X, NewPlace.Y);
          C := MovingList.GetLastCard; //  
          while C.isInternalMoving do Application.ProcessMessages; // ,       
        end
      else //       FREE_CELL
        begin
          SPEED_CARD := Round(SPEED_CARD * _nitro_speed);
          if nomer_active_st = 8 then MZ := -CARD_WIDTH div 4 else MZ := CARD_WIDTH div 4;  //      
          MoveCardWithDelay(MovingList.GetFirstCard, NewPlace.X + MZ, NewPlace.Y + CARD_HEIGHT div 3); //    ,    
          C := MovingList.GetLastCard; //  ,     
          while C.isInternalMoving do Application.ProcessMessages; // ,       
          for i := 0 to MovingList.ListOfCard.Count - 1 do MovingList.GetCard(i).fParent := nil; //   
          for i := 1 to 12 do {8+4}
            begin
              ListFC := ArrayStolb[i];
              if (ListFC.ListOfCard.Count = 0) and (ListFC <> NewLC) then //      ,  FreeCell !!!
                begin
                  MoveCardWithDelay(C, ListFC.LeftCoord, ListFC.TopCoord); //       FreeCell
                  C := MovingList.GetCard(MovingList.ListOfCard.IndexOf(C) - 1); //   ( ,     parent
                  if MovingList.ListOfCard.IndexOf(C) = 0 then Break;//    () ,        HomePlace
                end;
            end;
          for i := 0 to MovingList.ListOfCard.Count - 1 do //        
            begin
              C := MovingList.GetCard(i);
              MoveCardWithDelay(C, NewPlace.X, NewPlace.Y);
              C.fParent := NewLC;  //  
              NewPlace.Y := NewPlace.Y + _MinZazor;
            end;
          SPEED_CARD := Round(SPEED_CARD/_nitro_speed);          
        end;
    end; // end FREE_CELL (*)

  MovingList.ListOfCard.Clear;
  MovingList.Free;
end;

procedure TSoliter.SetText(txt: string);
begin
  if isAutomaticMode then Exit;
  TEXT_info := txt;
  if txt <> '' then
    LeftTimeToHideText := _TimeShowText
  else
    LeftTimeToHideText := 0;
end;

function TSoliter.GetText: string;
begin
  if LeftTimeToHideText <= 0 then TEXT_info := '  : ' + IntToStr(GetCountMoves);
  Result := TEXT_info;
end;

procedure TSoliter.TryAutomaticMove;

  function TryMoveCard(MovedCard: TCard): Boolean;
  var
    i: Integer;
  begin
    Result := False;
    if (MovedCard = nil) or (MovedCard = HeWasInHome) then Exit;
    for i := 13 to 16 {home} do
      if CanDropCard(MovedCard, ArrayStolb[i]) then //    
        begin //  
          GoToTheNewPlace(MovedCard, ArrayStolb[i]);
          Result := True;
          Break;
        end;
  end;

var
  i,j: Integer;
  MovedCard: TCard;
  CC,MC: Integer;
  ProcentAutoMove: Extended;
begin
  isAutomaticMode := True;
  ProcentAutoMove := 0;
  repeat
    CC := 0;
    MC := 0;
    for j := 1 to 3 do //   ,    - 
    for i := 1 to 8+4 do
      begin
        MovedCard := ArrayStolb[i].GetLastCard;
        if MovedCard <> nil then Inc(CC); //  
        if TryMoveCard(MovedCard) then Inc(MC); //   
      end;
    if CC = 0 then Break;  //      (   ...)
    ProcentAutoMove := 100*MC / CC;
  until ProcentAutoMove < _PROCENT_AUTO_MOVE; //    _PROCENT_AUTO_MOVE    
  isAutomaticMode := False;
  HeWasInHome := nil; //     
end;

function TSoliter.GetIndexOfListCard(List: TListCard): Integer;
var
  i: Integer;
begin
  Result := 0;
  for i := 1 to 16 do
    if List = ArrayStolb[i] then
      begin
        Result := i;
        Break;
      end;
end;

function TSoliter.GetIndexOfCard(C: TCard): Integer;
var
  i: Integer;
begin
  Result := 0;
  for i := 1 to 52 do
    if C = ArrayCards[i] then
      begin
        Result := i;
        Break;
      end;
end;

function IntToStr2(val: Integer): string;
begin
  Result := IntToStr(val);
  if Length(Result) < 2 then
    Result := '0' + Result;
end;

procedure TSoliter.SaveMoveToUndoList(Card: TCard);
var
  UndoString: string;
begin
  UndoString := IntToStr2(GetIndexOfCard(Card)) + IntToStr2(GetIndexOfListCard(Card.fParent));
  UndoList.Add(UndoString);
end;

function TSoliter.CanUndo: Boolean;
begin
  Result := UndoList.Count > 0;
end;

procedure TSoliter.DoUndo;
var
  UndoString: string;
  NomCard, NomList: Integer;
begin
  if UndoList.Count = 0 then Exit;
  UndoString := UndoList[UndoList.Count - 1];
  UndoList.Delete(UndoList.Count - 1);
  NomCard := StrToInt(Copy(UndoString, 1,2));
  NomList := StrToInt(Copy(UndoString, 3,2));
  isUndoMode := True; //    
  GoToTheNewPlace(ArrayCards[NomCard], ArrayStolb[NomList]); //  
  isUndoMode := False; //    
  InitVariantsMove;
end;

procedure TSoliter.InitVariantsMove;

  function Init_VM_MoveOnlyLastCard(Card: TCard; AimIND, LEVEL: Integer): Boolean;
  var
    MoveString: string;
  begin
    Result := False;
    if Card = ArrayStolb[AimIND].GetLastCard then Exit; //     
    Result := CanDropCard(Card, ArrayStolb[AimIND]);
    if Result then
      begin
        MoveString := IntToStr2(GetIndexOfCard(Card)) + IntToStr2(AimIND) + IntToStr(LEVEL);
        VariantsMove.Add(MoveString);
      end;
  end;

  procedure Init_VM_MoveCardsFromListCardToListCard(Card: TCard; AimIND, CountFreeCell, LEVEL: Integer);
  var
    AimList: TListCard;
    HeightStolb: Integer;
  begin
    if Card = nil then Exit;
    AimList := ArrayStolb[Aimind];
    if Card = AimList.GetLastCard then Exit; //     
    if AimList.ListOfCard.Count = 0 then Dec(CountFreeCell); //     
    HeightStolb := 0;
    repeat
      Inc(HeightStolb);
      if HeightStolb > CountFreeCell + 1 then Break;
      if CanStartMoveCard(Card) then //     
        Init_VM_MoveOnlyLastCard(Card, AimIND, LEVEL) //     
      else
        Break; //     ,     
      Card := Card.GetPrevCard //   
    until Card = nil //    Card  nil'
  end;

  var
    i,j, CFC: Integer;

begin
  isAutomaticMode := True; //      SetText
  VariantsMove.Clear;
  LastVariantMove := '';

  for i := 01 to 12 do //    free cell'
  for j := 13 to 16 do //  
    Init_VM_MoveOnlyLastCard(ArrayStolb[i].GetLastCard, j, 1); //      

  for i := 09 to 12 do //  free cell'
  for j := 01 to 08 do //  
    Init_VM_MoveOnlyLastCard(ArrayStolb[i].GetLastCard, j, 2); //    FreeCell    

  CFC := 0;
  for i := 1 to 12 do
    if ArrayStolb[i].GetFirstCard = nil then Inc(CFC); //  -  
  for i := 1 to 8 do //  
  for j := 1 to 8 do //  
    Init_VM_MoveCardsFromListCardToListCard(ArrayStolb[i].GetLastCard,j, CFC, 3); //       ( )

  for i := 01 to 08 do //  
  for j := 09 to 12 do //  free cell
    Init_VM_MoveOnlyLastCard(ArrayStolb[i].GetLastCard, j, 4); //          FreeCell

  isAutomaticMode := False;
end;

function TSoliter.GetCountMoves: Integer;
begin
  Result := VariantsMove.Count;
end;

procedure DrawOreolRect(Canvas: TCanvas; Left, Top, Right, Bottom: Integer; ColorRect: TColor); //   ,  _SizeOreol
var
  r,g,b: Byte;
  RectSh: TRect;
  x,y, dx, dw: Integer;
  p: PRGB_record;
  shd1, shd2: Real;
  Y1,Y2: Integer;
begin
  if Right - Left > _WidthCard then Right := Left + _WidthCard; //   ,        (  "  ")

  r := (ColorRect and $0000FF);
  g := (ColorRect and $00FF00) shr 8;
  b := (ColorRect and $FF0000) shr 16;
  RideTheLighting.Height := Bottom - Top + 2*_SizeOreol;

  RectSh := Rect(Left-_SizeOreol, Top-_SizeOreol, Left+_WidthCard+_SizeOreol, Bottom+_SizeOreol);
  RideTheLighting.Canvas.CopyRect(RideTheLighting.Canvas.ClipRect, Canvas, RectSh);
  Y1 := _SizeOreol + _RadiusCard -1;
  Y2 := RideTheLighting.Height - Y1 -1;
  
  dw := RideTheLighting.Width - (Right - Left) - 2*_SizeOreol;

  for y := 0 to RideTheLighting.Height - 1 do
    begin
      p := RideTheLighting.ScanLine[y];
      dx := 0;
      for x := 0 to Right - Left + 2*_SizeOreol - 1 do
        begin
          if (dx = 0) and (x > _SizeOreol+_RadiusCard) then dx := dw;
          if y < Y1 then //  
            shd1 := AlphaBlendLighting[x+dx,y]
          else
            if y > Y2 then //   !
              shd1 := AlphaBlendLighting[x+dx,ablY-y+Y2-1]
            else //  
              shd1 := AlphaBlendLighting[x+dx,Y1];
          shd2 := 1 - shd1;
          p^.R := MixColors(p^.R, r, shd1, shd2);
          p^.G := MixColors(p^.G, g, shd1, shd2);
          p^.B := MixColors(p^.B, b, shd1, shd2);
          Inc(p);
        end;
    end;
  Canvas.Draw(RectSh.Left, RectSh.Top, RideTheLighting);
end;

procedure TSoliter.ShowVariantHoda;
var
  VariantHoda: string;
  Card, onC: TCard;
  NewPlace: TListCard;
  NomCard, NomList, variant, i, level, lev: Integer;
  TempList: TStringList;
begin
  if VariantsMove.Count = 0 then Exit;
  if is_moving or isAutomaticMode or isUndoMode or move_with_delay then Exit; //    
// variant      Random[VariantsMove.Count],    
  TempList := TStringList.Create;
  Randomize;
  repeat
    for i := 0 to VariantsMove.Count div 2 do //    50%   ,      ,    
      TempList.Add(VariantsMove[Random(VariantsMove.Count)]);
    variant := 0;
    level := 5; //  level 1..4
    for i := 0 to TempList.Count - 1 do //        LEVEL (" ")
      begin
        VariantHoda := TempList[i];
        lev := StrToInt(Copy(VariantHoda, 5,1));
        if lev < level then
          begin
            level := lev;
            variant := i; //   TempList  " ", 
          end;
      end;
    VariantHoda := TempList[variant]; //    
    TempList.Clear;    
  until (VariantHoda <> LastVariantMove) or (VariantsMove.Count = 1); //           1!
  TempList.Free;
  LastVariantMove := VariantHoda;
//   VariantHoda,  :
  NomCard := StrToInt(Copy(VariantHoda, 1,2));
  NomList := StrToInt(Copy(VariantHoda, 3,2));
  Card := ArrayCards[NomCard];
  NewPlace := ArrayStolb[NomList];
  onC := NewPlace.GetLastCard;

  DrawGame; //   :
  if onC <> nil then
    DrawOreolRect(Canvas, OnC.Left, OnC.Top, OnC.Left + CARD_WIDTH, OnC.Top + CARD_HEIGHT, COLOR_AIM)
  else
    DrawOreolRect(Canvas, NewPlace.LeftCoord, NewPlace.TopCoord, NewPlace.LeftCoord + CARD_WIDTH, NewPlace.TopCoord + CARD_HEIGHT, COLOR_AIM);
  DrawOreolRect(Canvas, Card.Left, Card.Top, Card.fParent.GetLastCard.Left + CARD_WIDTH, Card.fParent.GetLastCard.Top + CARD_HEIGHT, COLOR_FROM);    
end;

function GetTypeOfColor(r,g,b: Byte): TTypeColor;
const
  porog = 128; //    ,  "="  ">",   !
begin
  Result := tcPicture;
  if (r=g) and (g=b) and (r<porog) then
    Result := tcMask
  else
    if (r=b) and (b<porog) then {   , ..     !!!}
      Result := tcShadow;
end;

procedure TransormColor(r,g,b: Byte; var RR,GG,BB: Byte);
var
  v: Real;
begin
  case GetTypeOfColor(r,g,b) of
    tcMask: begin
              RR := Round((255-RR)*r/255 + RR);
              GG := Round((255-GG)*g/255 + GG);
              BB := Round((255-BB)*b/255 + BB);
            end;
    tcShadow: begin
                if g <> 0 then v := 1-(g-r)/g else v := 0;
                RR := Round(((255-RR)*g/255 + RR) * v);
                GG := Round(((255-GG)*g/255 + GG) * v);
                BB := Round(((255-BB)*g/255 + BB) * v);
              end;
  else
    begin
      RR := r;
      GG := g;
      BB := b;
    end;
  end;
end;

procedure TransformBitMap(Color: TColor; BITMAP: TBitMap);
var
  x,y: Integer;
  P: PRGB_record;
  r,g,b: Byte;
  rr,gg,bb: Byte;  
begin
  r := (Color and $0000FF);
  g := (Color and $00FF00) shr 8;
  b := (Color and $FF0000) shr 16;
  for y := 0 to BITMAP.Height - 1 do
    begin
      P := BITMAP.ScanLine[y];
      for x := 0 to BITMAP.Width - 1 do
        begin
          rr := r; gg := g; bb := b;
          TransormColor(P.R,P.G,P.B, rr,gg,bb); 
          P.R := rr;
          P.G := gg;
          P.B := bb;
          Inc(P);
        end;
    end;
end;

procedure TSoliter.LoadParamFromIniFile;
var
  ini: TIniFile;
const
  MAIN = 'MAIN';
begin
  ini := TIniFile.Create(APPLICATION_PATH + 'Soliter.ini');
  Parent.Left := ini.ReadInteger(MAIN, 'Left', 0);
  Parent.Top := ini.ReadInteger(MAIN, 'Top', 0);  
  Parent.Width := ini.ReadInteger(MAIN, 'Width', 0);
  Parent.Height := ini.ReadInteger(MAIN, 'Height', 0);
  if ini.ReadBool(MAIN, 'isMaximized', False) then TForm(Parent).WindowState := wsMaximized;
  NEED_SHOW_FREE_CELL := ini.ReadBool(MAIN, 'NeedFreeCell', TRUE);
  COLOR_FROM := ini.ReadInteger(MAIN, 'Color1',   clAqua);
  COLOR_AIM  := ini.ReadInteger(MAIN, 'Color2', clYellow);
  FON_COLOR  := ini.ReadInteger(MAIN, 'FonColor', clTeal);
  SPEED_CARD := ini.ReadInteger(MAIN, 'Speed', SPEED_CARD);
  ChangeConturColor;
  ini.Free;
end;

procedure TSoliter.SaveParamToIniFile;
var
  ini: TIniFile;
  FormPlacement: TWindowPlacement;  
const
  MAIN = 'MAIN';
begin
  FormPlacement.length := SizeOf(TWindowPlacement);
  GetWindowPlacement(TForm(Parent).Handle, @FormPlacement); //  ""    (     )
  ini := TIniFile.Create(APPLICATION_PATH + 'Soliter.ini');
  ini.WriteInteger(MAIN, 'Left', FormPlacement.rcNormalPosition.Left);
  ini.WriteInteger(MAIN, 'Top', FormPlacement.rcNormalPosition.Top);
  ini.WriteInteger(MAIN, 'Width', FormPlacement.rcNormalPosition.Right - FormPlacement.rcNormalPosition.Left);
  ini.WriteInteger(MAIN, 'Height', FormPlacement.rcNormalPosition.Bottom - FormPlacement.rcNormalPosition.Top);
  ini.WriteBool(MAIN, 'isMaximized', TForm(Parent).WindowState = wsMaximized);
  ini.WriteBool(MAIN, 'NeedFreeCell', NEED_SHOW_FREE_CELL);
  ini.WriteInteger(MAIN, 'Color1',   COLOR_FROM);
  ini.WriteInteger(MAIN, 'Color2', COLOR_AIM);
  ini.WriteInteger(MAIN, 'FonColor', FON_COLOR);
  ini.WriteInteger(MAIN, 'Speed', SPEED_CARD);  
  ini.Free;
end;

procedure TSoliter.ChangeConturColor;
var
  i: Integer;
begin
  for i := 1 to 52 do ArrayCards[i].RedrawKontur;
end;

procedure TSoliter.RecalcNewSizeGame;
var
  i: Integer;
  sw,sh,SCALE: Extended;
  isChangeRazmOfCard: Boolean;

  function ScaleWidth: Extended;
  begin
    Result := Parent.ClientWidth/_WidthCard/10; 
  end;

  function ScaleHeight: Extended;
  begin
    Result := Parent.ClientHeight/_HeightCard/5.8;   
  end;
  
  procedure RecalcGeomParam;
  var
    K: Extended;
  begin
    K := sqrt(Width/CARD_WIDTH/10);
    if K < 1 then K := 1;
    if K > 1.3 then K := 1.3;
    _MinZazor := CARD_HEIGHT div 6;
    _MaxZazor := CARD_HEIGHT div 3;
    _ShagKart := Round(CARD_WIDTH * 1.2*K);
    _TopTopZazor := CARD_HEIGHT div 7;    
    _TopZazor := CARD_HEIGHT + 2*_TopTopZazor;
    _LeftZazor := (Width - (_ShagKart*7 + CARD_WIDTH)) div 2;
    _Left1Zazor := _LeftZazor;
    _TopShagKart := Round(CARD_WIDTH * 1.1*K);    
    _Left2Zazor := _LeftZazor + _ShagKart*7 - _TopShagKart*3;
  end;

  function SetScaleToCard(Scale: Extended): Boolean;  //   , TRUE   
  var
    NewW, NewH: Integer;
  begin
    Result := False;
    NewW := Round(_WidthCard * Scale);
    NewH := Round(_HeightCard * Scale);
    if (NewW <> CARD_WIDTH) or (NewH <> CARD_HEIGHT) then
      begin
        CARD_WIDTH := NewW;
        CARD_HEIGHT := NewH;
        Result := True;
      end;
  end;
                 
  procedure MoveCardOnRealPlace;
  var
    IND: Integer;
    ListCard: TListCard;
    FC, NC: TCard;
  begin
     for IND := 1 to 16 do
       begin
         ListCard := ArrayStolb[IND];
         case ListCard.TypeList of
            tKoloda: begin
                       ListCard.LeftCoord := _LeftZazor + (IND-1)*_ShagKart;
                       ListCard.TopCoord := _TopZazor;
                     end;
          tFreeCell: begin
                       ListCard.LeftCoord := _Left1Zazor + (IND-9)*_TopShagKart;
                       ListCard.TopCoord := _TopTopZazor;
                     end;
          tHomeCell: begin
                       ListCard.LeftCoord := _Left2Zazor + (IND-13)*_TopShagKart;
                       ListCard.TopCoord := _TopTopZazor;
                     end;        
         end;            
         FC := ListCard.GetFirstCard;
         if FC <> nil then
           begin
             FC.Left := ListCard.LeftCoord;
             FC.Top := ListCard.TopCoord;
             NC := FC.GetNextCard;       
             while NC <> nil do
               begin
                 NC.Left := FC.Left;
                 if ListCard.TypeList = tKoloda then NC.Top := FC.Top + _MinZazor else NC.Top := FC.Top;
                 FC := NC;
                 NC := FC.GetNextCard;                                             
               end;
           end;
       end;
  end;

  procedure DrawNewFon;
  var
    i: Integer;
    x,y: Integer;
    TempPic, TempPic2: TBitmap;
  begin
    BIG_FON_FOR_STRETCH.Width := Round(Width * SMOOTH / SCALE);
    BIG_FON_FOR_STRETCH.Height := Round(Height * SMOOTH / SCALE);    
    with BIG_FON_FOR_STRETCH.Canvas do
      begin
        Brush.Color := GIRL_PICTURE.Canvas.Pixels[0,0];
        FillRect(ClipRect);
        for i := 1 to 16 do
          begin
            X := Round(ArrayStolb[i].LeftCoord*SMOOTH/SCALE);
            Y := Round(ArrayStolb[i].TopCoord*SMOOTH/SCALE);
            if ArrayStolb[i].TypeList = tKoloda then
              Draw(X,Y, PLACE_DOWN_ROUND)
            else
              Draw(X,Y, PLACE_UP_ROUND);            
          end;
      end;
    BIG_FON_FOR_STRETCH.Canvas.Draw((BIG_FON_FOR_STRETCH.Width - GIRL_PICTURE.Width) div 2,Round(_HeightCard*3.2*SMOOTH),GIRL_PICTURE);
    FON_BITMAP.Width := 0; //  ...
    StretchPicture(FON_BITMAP, BIG_FON_FOR_STRETCH, Width, Height);
    TransformBitMap(FON_COLOR, FON_BITMAP);
    TForm(Parent).Color := FON_BITMAP.Canvas.Pixels[0,0];
    TempPic := TBitmap.Create;
    TempPic2 := TBitmap.Create;
    TempPic.Assign(LOGO);
    TransformBitMap(FON_COLOR, TempPic);
    i := Round(CARD_WIDTH * 0.74);
    StretchPicture(TempPic2, TempPic, i, i);
    FON_BITMAP.Canvas.Draw((Width - i) div 2, _TopTopZazor, TempPic2);
    TempPic.Free;    
    TempPic2.Free;
  end;
  
begin
  sw := ScaleWidth;
  sh := ScaleHeight;
  if sw < sh then SCALE := sw else SCALE := sh;
  if SCALE > 1 then SCALE := 1;
  if SCALE < 0.3 then SCALE := 0.3;
  isChangeRazmOfCard :=  SetScaleToCard(SCALE);
  RecalcGeomParam;
  MoveCardOnRealPlace;
  if isChangeRazmOfCard then //      
    begin
      for i := 1 to 52 do ArrayCards[i].STRETCH_CARD;
      PLACE_UP_ROUND.Assign(PLACE_UP);
      RoundPic(PLACE_UP_ROUND, Round((_RadiusCard+1) * SMOOTH / SCALE));
      PLACE_DOWN_ROUND.Assign(PLACE_DOWN);
      RoundPic(PLACE_DOWN_ROUND, Round((_RadiusCard+1) * SMOOTH / SCALE));
    end;
  DrawNewFon;
  DrawGame;
end;

//\\ ///\\\ ////\\\\ /////\\\\\ //////\\\\\\

function Min(a1,a2: Integer): Integer;
begin
  if a1 < a2 then Result := a1 else Result := a2;
end;

function Max(a1,a2: Integer): Integer;
begin
  if a1 > a2 then Result := a1 else Result := a2;
end;

function CalcFadeLinear(EndLevel, dist: Real; SizeFade: Integer): Real;
begin
  if dist < 0 then dist := 0; //   
  Result := (dist*(1 - EndLevel) + EndLevel*SizeFade)/SizeFade;
end;

function CalcFadeNonLinear(dist: Real; SizeFade: Integer): Real;
begin
  if dist < 0 then dist := 0; //   
  Result := Sqrt(Sqrt(abs(dist-0.5)/SizeFade)); //  ,     !
end;

procedure DrawGradientCircle(BitMap: TBitMap; Xcenter,Ycenter, IntRadius, SizeDraw: Integer; EndLevel: Real; isLinear: Boolean);
var
  X1,X2,Y1,Y2,x,y: Integer;
  dist, sh: Real;
  col: Byte;
begin
  X1 := Max(0, Xcenter - IntRadius - SizeDraw);
  X2 := Min(BitMap.Width - 1, Xcenter + IntRadius + SizeDraw);
  Y1 := Max(0, Ycenter - IntRadius - SizeDraw);
  Y2 := Min(BitMap.Height - 1, Ycenter + IntRadius + SizeDraw);
  with BitMap.Canvas do
    for x := X1 to X2 do
    for y := Y1 to Y2 do
      begin
        dist := sqrt(sqr(Xcenter-x)+sqr(Ycenter-y));
        if (dist < IntRadius - 1) or (dist > IntRadius + SizeDraw) then
          Pixels[x,y] := clWhite
        else
          begin
            dist := dist - IntRadius;
            if isLinear then
              sh := CalcFadeLinear(EndLevel, dist, SizeDraw)
            else
              sh := CalcFadeNonLinear(dist, SizeDraw);
            col := FadeToBlack(255, sh);
            Pixels[x,y] := RGB(col,col,col);
          end;
      end;
end;

var
  BM_ALFA: TBitmap; //  
  shd: Extended;
  i, j: Integer;
  col: Byte;

  procedure GradientRect(FromRGB, ToRGB: TColor; Canvas: TCanvas);    //      
  var
    RGBFrom : array[0..2] of Byte; { from RGB values }
    RGBDiff : array[0..2] of integer; { difference of from/to RGB values }
    ColorBand : TRect; { color band rectangular coordinates }
    I : Integer; { color band index }
    R : Byte; { a color band's R value }
    G : Byte; { a color band's G value }
    B : Byte; { a color band's B value }
  begin
    { extract from RGB values}
    RGBFrom[0] := GetRValue (ColorToRGB (FromRGB));
    RGBFrom[1] := GetGValue (ColorToRGB (FromRGB));
    RGBFrom[2] := GetBValue (ColorToRGB (FromRGB));
    { calculate difference of from and to RGB values}
    RGBDiff[0] := GetRValue (ColorToRGB (ToRGB)) - RGBFrom[0];
    RGBDiff[1] := GetGValue (ColorToRGB (ToRGB)) - RGBFrom[1];
    RGBDiff[2] := GetBValue (ColorToRGB (ToRGB)) - RGBFrom[2];

    { set pen sytle and mode}
    Canvas.Pen.Style := psSolid;
    Canvas.Pen.Mode := pmCopy;

    { set color band's left and right coordinates}
    ColorBand.Left := 0;
    ColorBand.Right:= canvas.ClipRect.Right-Canvas.ClipRect.Left;

    for I := 0 to $ff do
    begin
      { calculate color band's top and bottom coordinates}
      ColorBand.Top := MulDiv (I , canvas.ClipRect.Bottom-Canvas.ClipRect.Top, $100);
      ColorBand.Bottom := MulDiv (I + 1,canvas.ClipRect.Bottom-Canvas.ClipRect.Top, $100);
      { calculate color band color}
      R := RGBFrom[0] + MulDiv (I, RGBDiff[0], $ff);
      G := RGBFrom[1] + MulDiv (I, RGBDiff[1], $ff);
      B := RGBFrom[2] + MulDiv (I, RGBDiff[2], $ff);
      { select brush and paint color band}
      Canvas.Brush.Color := RGB (R, G, B);
      Canvas.FillRect (ColorBand);
    end;
  end;
    
  procedure DrawPlace(PlacePic: TBitmap; isUp: Boolean);
  var
    W,H,T: Integer;
  begin
    W := Round(_WidthCard*SMOOTH);  //   SMOOTH       
    H := Round(_HeightCard*SMOOTH);
    if not isUp then H := Round(H*0.8); //    
    PlacePic.Width := W;
    PlacePic.Height := H;
    if not isUp then
      begin
        GradientRect($333333, GIRL_PICTURE.Canvas.Pixels[0,0], PlacePic.Canvas);      
        T := W div 10;       
        PlacePic.Canvas.Brush.Color := TRANSPARENT_COLOR;
        PlacePic.Canvas.FillRect(Rect(T, T, W-T, H));
      end
    else
      GradientRect($222222, $444444, PlacePic.Canvas);
  end;

initialization

  APPLICATION_PATH := ExtractFilePath(Application.ExeName);

  BM_ALFA := TBitmap.Create; //        
  BM_ALFA.PixelFormat := pf24bit;

  // :
  BM_ALFA.Width := _WidthCard + _SizeShadow;
  BM_ALFA.Height := _HeightCard + _SizeShadow;

  DrawGradientCircle(BM_ALFA, BM_ALFA.Width - _RadiusCard - _SizeShadow, BM_ALFA.Height - _RadiusCard - _SizeShadow, _RadiusCard, _SizeShadow, _ShadowLevel, TRUE);
  DrawGradientCircle(BM_ALFA, BM_ALFA.Width - _RadiusCard - _SizeShadow, 2*_SizeShadow +_RadiusCard - _DeltaShadow, _RadiusCard, _SizeShadow, _ShadowLevel, TRUE);
  DrawGradientCircle(BM_ALFA, 2*_SizeShadow + _RadiusCard - _DeltaShadow, BM_ALFA.Height - _RadiusCard - _SizeShadow, _RadiusCard, _SizeShadow, _ShadowLevel, TRUE);
  for i := 0 to _SizeShadow - 1 do
    with BM_ALFA.Canvas do
      begin
        shd := CalcFadeLinear(_ShadowLevel, i, _SizeShadow);
        col := FadeToBlack(255, shd);
        Pen.Color := RGB(col,col,col);
        MoveTo(BM_ALFA.Width - _SizeShadow + i, BM_ALFA.Height - _RadiusCard - _SizeShadow);
        LineTo(BM_ALFA.Width - _SizeShadow + i, 2*_SizeShadow +_RadiusCard - _DeltaShadow);
        MoveTo(2*_SizeShadow + _RadiusCard - _DeltaShadow, BM_ALFA.Height - _SizeShadow + i);
        LineTo(BM_ALFA.Width - _RadiusCard - _SizeShadow, BM_ALFA.Height - _SizeShadow + i);
      end;

  BM_ALFA.Canvas.Pen.Color := clWhite;
  BM_ALFA.Canvas.RoundRect(0,0, BM_ALFA.Width-_SizeShadow,BM_ALFA.Height-_SizeShadow, _RadiusCard*2, _RadiusCard*2);

//  BM_ALFA.SaveToFile('Shadow.bmp');

  for i := 0 to absX1-1 do
  for j := 0 to absY1-1 do
    AlphaBlendShadow1[i,j] := Byte(BM_ALFA.Canvas.Pixels[i + _WidthCard, j + _SizeShadow - _DeltaShadow])/255;

  for i := 0 to absX2-1 do
  for j := 0 to absY2-1 do
    AlphaBlendShadow2[i,j] := Byte(BM_ALFA.Canvas.Pixels[i + _SizeShadow - _DeltaShadow, j + _HeightCard - _RadiusCard])/255;

  // :
  BM_ALFA.Width := ablX;
  BM_ALFA.Height := ablY;

  DrawGradientCircle(BM_ALFA, _SizeOreol + _RadiusCard -1, BM_ALFA.Height -1, _RadiusCard, _SizeOreol, 0, FALSE);
  DrawGradientCircle(BM_ALFA, BM_ALFA.Width - _SizeOreol - _RadiusCard, BM_ALFA.Height -1, _RadiusCard, _SizeOreol, 0, FALSE);
  for i := 0 to _SizeOreol - 1 do
    with BM_ALFA.Canvas do
      begin
        shd := CalcFadeNonLinear(i, _SizeOreol);
        col := FadeToBlack(255, shd);
        Pen.Color := RGB(col,col,col);
        MoveTo(_SizeOreol + _RadiusCard, BM_ALFA.Height - _RadiusCard - i -1);
        LineTo(BM_ALFA.Width - _SizeOreol - _RadiusCard, BM_ALFA.Height - _RadiusCard - i -1);
      end;

  BM_ALFA.Canvas.Pen.Color := clWhite;
  BM_ALFA.Canvas.RoundRect(_SizeOreol, BM_ALFA.Height - _RadiusCard,
                           BM_ALFA.Width - _SizeOreol, BM_ALFA.Height + _RadiusCard,
                           _RadiusCard*2, _RadiusCard*2);

//  BM_ALFA.SaveToFile('Oreol.bmp');

  for i := 0 to ablX-1 do
  for j := 0 to ablY-1 do
    AlphaBlendLighting[i,j] := Byte(BM_ALFA.Canvas.Pixels[i,j])/255;

  BM_ALFA.Free;

  GIRL_PICTURE := TBitmap.Create;
  GIRL_PICTURE.LoadFromFile(APPLICATION_PATH + '\pic\GIRL.bmp');  
  
  LOGO := TBitmap.Create;
  LOGO.LoadFromFile(APPLICATION_PATH + '\pic\LOGO.bmp');  

  Shadow1 := TBitmap.Create;
  Shadow2 := TBitmap.Create;

  Shadow1.PixelFormat := pf24bit;
  Shadow2.PixelFormat := pf24bit;

  Shadow1.Width := absX1;
  Shadow1.Height := absY1;

  Shadow2.Width := absX2;
  Shadow2.Height := absY2;

  move_with_delay := False;

  RideTheLighting := TBitmap.Create;
  RideTheLighting.PixelFormat := pf24bit;
  RideTheLighting.Width := ablX;
  RideTheLighting.Height := 0; //  

  CARDS_BITMAP := TBitmap.Create;
  CARDS_BITMAP.LoadFromFile(APPLICATION_PATH + '\pic\CARDS_VISTA.bmp');

  FON_BITMAP := TBitmap.Create;
  FON_BITMAP.PixelFormat := pf24bit;

  PLACE_UP := TBitmap.Create;
  DrawPlace(PLACE_UP, True);

  PLACE_DOWN := TBitmap.Create;
  DrawPlace(PLACE_DOWN, False);  

  PLACE_UP_ROUND := TBitmap.Create;
  PLACE_DOWN_ROUND := TBitmap.Create;

  BIG_FON_FOR_STRETCH := TBitmap.Create;

finalization

  Shadow1.Free;
  Shadow2.Free;
  RideTheLighting.Free;
  
  CARDS_BITMAP.Free;
  FON_BITMAP.Free;
  PLACE_UP.Free;
  PLACE_DOWN.Free;

  PLACE_UP_ROUND.Free;
  PLACE_DOWN_ROUND.Free;

  BIG_FON_FOR_STRETCH.Free;

  GIRL_PICTURE.Free;
  LOGO.Free;
  
end.
